#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2024 The meson-python developers
#
# SPDX-License-Identifier: MIT

import os
import sys

outdir = os.path.join(sys.argv[1], "generated")
os.makedirs(outdir, exist_ok=True)

for name in "one.py", "two.py":
    with open(os.path.join(outdir, name), "w"):
        pass



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import contextlib
import os
import os.path
import pathlib
import re
import shutil
import subprocess
import sys
import sysconfig
import tempfile
import warnings

from venv import EnvBuilder

import packaging.metadata
import packaging.version
import pytest

import hatchling.build


@contextlib.contextmanager
def chdir(path):
    """Context manager helper to change the current working directory -- cd."""
    old_cwd = os.getcwd()
    os.chdir(os.fspath(path))
    try:
        yield path
    finally:
        os.chdir(old_cwd)


def metadata(data):
    meta, other = packaging.metadata.parse_email(data)
    assert not other
    return meta


def adjust_packaging_platform_tag(platform: str) -> str:
    if platform.startswith(("manylinux", "musllinux")):
        # The packaging module generates overly specific platforms tags on
        # Linux.  The platforms tags on Linux evolved over time.
        # meson-python uses more relaxed platform tags to maintain
        # compatibility with old wheel installation tools.  The relaxed
        # platform tags match the ones generated by the wheel package.
        # https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/
        return re.sub(
            r"^(many|musl)linux(1|2010|2014|_\d+_\d+)_(.*)$", r"linux_\3", platform
        )
    if platform.startswith("macosx"):
        # Python built with older macOS SDK on macOS 11, reports an
        # unexising macOS 10.16 version instead of the real version.
        # The packaging module introduced a workaround in version
        # 22.0.  Too maintain compatibility with older packaging
        # releases we don't implement it.  Reconcile this.
        from platform import mac_ver

        version = tuple(map(int, mac_ver()[0].split(".")))[:2]
        if version == (10, 16):
            return re.sub(r"^macosx_\d+_\d+_(.*)$", r"macosx_10_16_\1", platform)
    return platform


package_dir = pathlib.Path(__file__).parent / "packages"


@contextlib.contextmanager
def in_git_repo_context(path=os.path.curdir):
    # Resist the tentation of using pathlib.Path here: it is not
    # supporded by subprocess in Python 3.7.
    path = os.path.abspath(path)
    shutil.rmtree(os.path.join(path, ".git"), ignore_errors=True)
    try:
        subprocess.run(["git", "init", "-b", "main", path], check=True)
        subprocess.run(
            ["git", "config", "user.email", "author@example.com"], cwd=path, check=True
        )
        subprocess.run(["git", "config", "user.name", "A U Thor"], cwd=path, check=True)
        subprocess.run(["git", "add", "*"], cwd=path, check=True)
        subprocess.run(["git", "commit", "-q", "-m", "Test"], cwd=path, check=True)
        yield
    finally:
        # PermissionError raised on Windows.
        with contextlib.suppress(PermissionError):
            shutil.rmtree(os.path.join(path, ".git"))


@pytest.fixture(scope="session")
def tmp_path_session(tmp_path_factory):
    return pathlib.Path(
        tempfile.mkdtemp(
            prefix="mesonpy-test-",
            dir=tmp_path_factory.mktemp("test"),
        )
    )


class VEnv(EnvBuilder):
    def __init__(self, env_dir):
        super().__init__(symlinks=True, with_pip=True)

        # This warning is mistakenly generated by CPython 3.11.0
        # https://github.com/python/cpython/pull/98743
        with warnings.catch_warnings():
            if sys.version_info[:3] == (3, 11, 0):
                warnings.filterwarnings(
                    "ignore",
                    "check_home argument is deprecated and ignored.",
                    DeprecationWarning,
                )
            self.create(env_dir)

        # Free-threaded Python 3.13 requires pip 24.1b1 or later.
        if sysconfig.get_config_var("Py_GIL_DISABLED"):
            # importlib.metadata is not available on Python 3.7 and
            # earlier, however no-gil builds are available only for
            # Python 3.13 and later.
            import importlib.metadata

            if packaging.version.Version(
                importlib.metadata.version("pip")
            ) < packaging.version.Version("24.1b1"):
                self.pip("install", "--upgrade", "pip >= 24.1b1")

    def ensure_directories(self, env_dir):
        context = super().ensure_directories(env_dir)
        # Store the path to the venv Python interpreter. There does
        # not seem to be a way to do this without subclassing.
        self.executable = context.env_exe
        return context

    def python(self, *args: str):
        return subprocess.check_output([self.executable, *args]).decode()

    def pip(self, *args: str):
        return self.python("-m", "pip", *args)


@pytest.fixture()
def venv(tmp_path_factory):
    path = pathlib.Path(tmp_path_factory.mktemp("mesonpy-test-venv"))
    return VEnv(path)


def generate_package_fixture(package):
    @pytest.fixture
    def fixture():
        with chdir(package_dir / package) as new_path:
            yield new_path

    return fixture


def generate_copyof_package_fixture(package):
    @pytest.fixture
    def fixture(tmp_path):
        shutil.copytree(package_dir / package, tmp_path / package)

        with chdir(tmp_path / package) as new_path:
            yield new_path

    return fixture


def generate_sdist_fixture(package):
    @pytest.fixture(scope="session")
    def fixture(tmp_path_session):
        with chdir(package_dir / package), in_git_repo_context():
            return tmp_path_session / hatchling.build.build_sdist(tmp_path_session)

    return fixture


def generate_wheel_fixture(package):
    @pytest.fixture(scope="session")
    def fixture(tmp_path_session):
        shutil.copytree(package_dir / package, tmp_path_session / package)

        with chdir(tmp_path_session / package), in_git_repo_context():
            return tmp_path_session / hatchling.build.build_wheel(tmp_path_session)

    return fixture


# def generate_editable_fixture(package):
#     @pytest.fixture(scope="session")
#     def fixture(tmp_path_session):
#         # shutil.rmtree(
#         #     package_dir / package / ".mesonpy" / "editable", ignore_errors=True
#         # )
#         with chdir(package_dir / package), in_git_repo_context():
#             return tmp_path_session / hatchling.build.build_editable(tmp_path_session)

#     return fixture


# inject {package,sdist,wheel}_* fixtures (https://github.com/pytest-dev/pytest/issues/2424)
for package in os.listdir(package_dir):
    normalized = package.replace("-", "_")
    globals()[f"package_{normalized}"] = generate_package_fixture(package)
    globals()[f"copyof_{normalized}"] = generate_copyof_package_fixture(package)
    globals()[f"sdist_{normalized}"] = generate_sdist_fixture(package)
    globals()[f"wheel_{normalized}"] = generate_wheel_fixture(package)
    # globals()[f"editable_{normalized}"] = generate_editable_fixture(package)


@pytest.fixture(autouse=True, scope="session")
def disable_pip_version_check():
    # Cannot use the 'monkeypatch' fixture because of scope mismatch.
    mpatch = pytest.MonkeyPatch()
    yield mpatch.setenv("PIP_DISABLE_PIP_VERSION_CHECK", "1")
    mpatch.undo()


@pytest.fixture(autouse=True, scope="session")
def cleanenv():
    # Cannot use the 'monkeypatch' fixture because of scope mismatch.
    mpatch = pytest.MonkeyPatch()
    # $MACOSX_DEPLOYMENT_TARGET affects the computation of the platform tag on macOS.
    yield mpatch.delenv("MACOSX_DEPLOYMENT_TARGET", raising=False)
    mpatch.undo()



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import os
import re
import shutil
import subprocess
import textwrap

from typing import List

import packaging.requirements
import pytest

import hatch_meson.plugin


@pytest.mark.parametrize(
    "ninja", [None, "1.8.1", "1.8.3"], ids=["noninja", "oldninja", "newninja"]
)
def test_get_requires_for_build_wheel(monkeypatch, package_pure, ninja):
    # the NINJA environment variable affects the ninja executable lookup and breaks the test
    monkeypatch.delenv("NINJA", raising=False)

    def which(prog: str) -> bool:
        if prog == "ninja":
            return ninja and "ninja"
        if prog in ("ninja-build", "samu"):
            return None
        # smoke check for the future if we add another usage
        raise AssertionError(f"Called with {prog}, tests not expecting that usage")

    subprocess_run = subprocess.run

    def run(
        cmd: List[str], *args: object, **kwargs: object
    ) -> subprocess.CompletedProcess:
        if cmd == ["ninja", "--version"]:
            return subprocess.CompletedProcess(cmd, 0, f"{ninja}\n", "")
        return subprocess_run(cmd, *args, **kwargs)

    monkeypatch.setattr(shutil, "which", which)
    monkeypatch.setattr(subprocess, "run", run)

    expected = set()

    if ninja is None or hatch_meson.plugin._parse_version_string(ninja) < (1, 8, 2):
        expected.add("ninja")

    requirements = hatch_meson.plugin._get_requires_for_build_wheel()

    # Check that the requirement strings are in the correct format.
    names = {packaging.requirements.Requirement(x).name for x in requirements}

    assert names == expected


@pytest.mark.parametrize("meson", [None, "meson"])
def test_get_meson_command(monkeypatch, meson):
    # The MESON environment variable affects the meson executable lookup and breaks the test.
    monkeypatch.delenv("MESON", raising=False)
    assert hatch_meson.plugin._get_meson_command(meson) == ["meson"]


def test_get_meson_command_bad_path(monkeypatch):
    # The MESON environment variable affects the meson executable lookup and breaks the test.
    monkeypatch.delenv("MESON", raising=False)
    with pytest.raises(
        hatch_meson.plugin.ConfigError,
        match=re.escape('meson executable "bad" not found'),
    ):
        hatch_meson.plugin._get_meson_command("bad")


def test_get_meson_command_bad_python_path(monkeypatch):
    # The MESON environment variable affects the meson executable lookup and breaks the test.
    monkeypatch.delenv("MESON", raising=False)
    with pytest.raises(
        hatch_meson.plugin.ConfigError,
        match=re.escape('Could not find the specified meson: "bad-python-path.py"'),
    ):
        hatch_meson.plugin._get_meson_command("bad-python-path.py")


def test_get_meson_command_wrong_version(monkeypatch, tmp_path):
    # The MESON environment variable affects the meson executable lookup and breaks the test.
    monkeypatch.delenv("MESON", raising=False)
    meson = tmp_path / "meson.py"
    meson.write_text(textwrap.dedent("""
        print('0.0.1')
    """))
    with pytest.raises(
        hatch_meson.plugin.ConfigError,
        match=r"Could not find meson version [0-9\.]+ or newer, found 0\.0\.1\.",
    ):
        hatch_meson.plugin._get_meson_command(os.fspath(meson))


def test_get_meson_command_error(monkeypatch, tmp_path):
    # The MESON environment variable affects the meson executable lookup and breaks the test.
    monkeypatch.delenv("MESON", raising=False)
    meson = tmp_path / "meson.py"
    meson.write_text(textwrap.dedent("""
        import sys
        print('Just testing', file=sys.stderr)
        sys.exit(1)
    """))
    with pytest.raises(
        hatch_meson.plugin.ConfigError,
        match=re.escape("Could not execute meson: Just testing"),
    ):
        hatch_meson.plugin._get_meson_command(os.fspath(meson))



# SPDX-FileCopyrightText: 2023 The meson-python developers
#
# SPDX-License-Identifier: MIT

import os
import subprocess

import pytest

import hatchling.build

ninja_ver_str = subprocess.run(
    ["ninja", "--version"], check=True, stdout=subprocess.PIPE, text=True
).stdout
NINJA_VERSION = tuple(map(int, ninja_ver_str.split(".")[:3]))


# Ninja 1.9 does not support the soruce^ syntax to specify a target.
@pytest.mark.skipif(NINJA_VERSION < (1, 10), reason="Ninja version too old")
@pytest.mark.parametrize(
    ("args", "expected"),
    [
        ([], True),
        (["-Dbuildtype=release"], True),
        (["-Dbuildtype=debug"], False),
    ],
    ids=["", "-Dbuildtype=release", "-Dbuildtype=debug"],
)
def test_ndebug(copyof_plat, tmp_path, args, expected):

    build_dir = copyof_plat / "build"
    config_args = {"setup-args": args, "build-dir": str(build_dir)}
    hatchling.build.build_wheel(tmp_path, config_args)
    command = subprocess.run(
        # Ask ninja what is the command that would be used to
        # compile a C source file (the trailing ^ is used to
        # specify the target that is the first output of the rule
        # containing the specified source file).
        [
            "ninja",
            "-C",
            os.fspath(build_dir),
            "-t",
            "commands",
            "../plat.c^",
        ],
        stdout=subprocess.PIPE,
        check=True,
    ).stdout
    assert (b"-DNDEBUG" in command) == expected



import dataclasses
import typing as T

# Cannot use `from __future__ import annotations` because validobj does not
# support annotations


@dataclasses.dataclass
class MesonArgs:
    """
    Contents of [tool.hatch.build.hooks.meson.args] items
    """

    setup: T.List[str] = dataclasses.field(default_factory=list)
    """Additional args to pass to ``meson setup``"""

    compile: T.List[str] = dataclasses.field(default_factory=list)
    """Additional args to pass to ninja or ``meson compile``"""

    install: T.List[str] = dataclasses.field(default_factory=list)
    """Arguments to ``meson install``"""

    # hatch-meson does not use `meson dist`


@dataclasses.dataclass
class HatchMesonConfig:
    """
    Contents of [tool.hatch.build.hooks.meson] items
    """

    meson: T.Optional[str] = None
    """Meson to use"""

    limited_api: bool = False
    """Build extension modules targeting the limited API"""

    args: MesonArgs = dataclasses.field(default_factory=lambda: MesonArgs())


@dataclasses.dataclass
class CmdlineConfig:
    """
    Contents of config settings passed in through a build tool
    """

    build_dir: T.Optional[str] = None
    editable_verbose: bool = False

    # dist_args: T.Union[None, str, T.List[str]] = None
    setup_args: T.Union[None, str, T.List[str]] = None
    compile_args: T.Union[None, str, T.List[str]] = None
    install_args: T.Union[None, str, T.List[str]] = None



# SPDX-FileCopyrightText: 2021 Filipe Laíns <lains@riseup.net>
# SPDX-FileCopyrightText: 2021 Quansight, LLC
# SPDX-FileCopyrightText: 2022 The meson-python developers
# SPDX-FileCopyrightText: 2024 The hatch-meson developers
#
# SPDX-License-Identifier: MIT
#
# Lots of code ripped from meson-python and some limited pieces inspired by the
# scikit-build-core hatchling plugin also
#

import argparse
import collections
import functools
import importlib.machinery
import importlib.metadata
import inspect
import json
import os
import pathlib
import platform
import re
import shutil
import subprocess
import sys
import sysconfig
import textwrap
import typing as T

from . import _tags
from .config import CmdlineConfig, HatchMesonConfig

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.version.source.plugin.interface import VersionSourceInterface

from validobj.validation import parse_input


# Terrible hack to get config_settings even though hatchling doesn't support it
# - https://github.com/pypa/hatch/issues/1072
def _get_cmdline_config_settings() -> T.Optional[T.Dict[str, T.Any]]:
    for frame_info in inspect.stack():
        frame = frame_info.frame
        module = inspect.getmodule(frame)
        if (
            module
            and module.__name__.startswith("hatchling.build")
            and "config_settings" in frame.f_locals
        ):
            return frame.f_locals["config_settings"]

    return None


_NINJA_REQUIRED_VERSION = "1.8.2"
_MESON_REQUIRED_VERSION = (
    "0.64.0"  # keep in sync with the version requirement in pyproject.toml
)


class Error(RuntimeError):
    def __str__(self) -> str:
        return str(self.args[0])


class ConfigError(Error):
    """Error in the backend configuration."""


class BuildError(Error):
    """Error when installing"""


_SUFFIXES = importlib.machinery.all_suffixes()
_EXTENSION_SUFFIX_REGEX = re.compile(r"^[^.]+\.(?:(?P<abi>[^.]+)\.)?(?:so|pyd|dll)$")
assert all(
    re.match(_EXTENSION_SUFFIX_REGEX, f"foo{x}")
    for x in importlib.machinery.EXTENSION_SUFFIXES
)

# Map Meson installation path placeholders to wheel installation paths.
# See https://docs.python.org/3/library/sysconfig.html#installation-paths
_INSTALLATION_PATH_MAP = {
    "{bindir}": "scripts",
    "{py_purelib}": "purelib",
    "{py_platlib}": "platlib",
    "{moduledir_shared}": "platlib",
    # hatchling doesn't support this, and nobody uses it anyways
    # "{includedir}": "headers",
    "{datadir}": "data",
    # custom location -- bundling not supported yet, and I'm not very comfortable
    # with how meson-python deals with this problem yet (but I don't have a good
    # rationale for this quite yet)
    # "{libdir}": "mesonpy-libs",
    # "{libdir_shared}": "mesonpy-libs",
}


def _pkgconf_binary_for_native_file() -> T.Optional[str]:
    try:
        dist = importlib.metadata.distribution("pkgconf")
    except importlib.metadata.PackageNotFoundError:
        return None

    for file in dist.files or ():
        if file.name == "pkgconf-pypi":
            return os.fspath(dist.locate_file(file))

    return None


def _map_to_wheel(
    sources: T.Dict[str, T.Dict[str, T.Any]],
) -> T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]]:
    """Map files to the wheel, organized by wheel installation directory."""
    wheel_files: T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]] = (
        collections.defaultdict(list)
    )
    # packages: T.Dict[str, str] = {}

    for key, group in sources.items():
        for src, target in group.items():
            destination = pathlib.Path(target["destination"])
            anchor = destination.parts[0]
            dst = pathlib.Path(*destination.parts[1:])

            path = _INSTALLATION_PATH_MAP.get(anchor)
            if path is None:
                raise BuildError(
                    f"Could not map installation path to an equivalent wheel directory: {str(destination)!r}"
                )

            # meson-python tries to support both purelib and platlib, but we're only going
            # to support one or the other. It'll be checked later.
            #
            # if path == "purelib" or path == "platlib":
            #     package = destination.parts[1]
            #     other = packages.setdefault(package, path)
            #     if other != path:
            #         this = os.fspath(pathlib.Path(path, *destination.parts[1:]))
            #         that = os.fspath(
            #             other
            #             / next(
            #                 d
            #                 for d, s in wheel_files[other]
            #                 if d.parts[0] == destination.parts[1]
            #             )
            #         )
            #         raise BuildError(
            #             f"The {package} package is split between {path} and {other}: "
            #             f'{this!r} and {that!r}, a "pure: false" argument may be missing in meson.build. '
            #             f"It is recommended to set it in \"import('python').find_installation()\""
            #         )

            if key == "install_subdirs" or key == "targets" and os.path.isdir(src):
                exclude_files = {
                    os.path.normpath(x) for x in target.get("exclude_files", [])
                }
                exclude_dirs = {
                    os.path.normpath(x) for x in target.get("exclude_dirs", [])
                }
                for root, dirnames, filenames in os.walk(src):
                    for name in dirnames.copy():
                        dirsrc = os.path.join(root, name)
                        relpath = os.path.relpath(dirsrc, src)
                        if relpath in exclude_dirs:
                            dirnames.remove(name)
                    # sort to process directories determninistically
                    dirnames.sort()
                    for name in sorted(filenames):
                        filesrc = os.path.join(root, name)
                        relpath = os.path.relpath(filesrc, src)
                        if relpath in exclude_files:
                            continue
                        filedst = dst / relpath
                        wheel_files[path].append((filedst, filesrc))
            else:
                wheel_files[path].append((dst, src))

    return wheel_files


def _is_native(fname) -> bool:
    """Check if file is a native file."""

    with open(fname, "rb") as f:
        if sys.platform == "darwin":
            return f.read(4) in (
                b"\xfe\xed\xfa\xce",  # 32-bit
                b"\xfe\xed\xfa\xcf",  # 64-bit
                b"\xcf\xfa\xed\xfe",  # arm64
                b"\xca\xfe\xba\xbe",  # universal / fat (same as java class so beware!)
            )
        elif sys.platform == "win32" or sys.platform == "cygwin":
            return f.read(2) == b"MZ"
        else:
            # Assume that any other platform uses ELF binaries.
            return f.read(4) == b"\x7fELF"  # ELF


def _install_is_pure(
    install_plan: T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]],
) -> bool:
    """Whether the wheel is architecture independent"""
    if install_plan["platlib"]:
        return False
    for _, src in install_plan["scripts"]:
        if _is_native(src):
            return False
    return True


def _compute_stable_abi(
    install_plan: T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]],
    limited_api: bool,
) -> T.Optional[str]:
    # PyPy supports the limited API but does not provide a stable
    # ABI, therefore extension modules using the limited API do
    # not use the stable ABI filename suffix and wheels should not
    # be tagged with the abi3 tag.
    if limited_api and "__pypy__" not in sys.builtin_module_names:
        # Verify stable ABI compatibility: examine files installed
        # in {platlib} that look like extension modules, and raise
        # an exception if any of them has a Python version
        # specific extension filename suffix ABI tag.
        for path, _ in install_plan["platlib"]:
            match = _EXTENSION_SUFFIX_REGEX.match(path.name)
            if match:
                abi = match.group("abi")
                if abi is not None and abi != "abi3":
                    raise BuildError(
                        f"The package declares compatibility with Python limited API but extension "
                        f"module {os.fspath(path)!r} is tagged for a specific Python version."
                    )
        return "abi3"
    return None


def _compute_tag(
    install_plan: T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]],
    limited_api: bool,
) -> _tags.Tag:
    """Wheel tags."""
    if _install_is_pure(install_plan):
        return _tags.Tag("py3", "none", "any")

    # Assume that all code installed in {platlib} is Python ABI dependent.
    if not install_plan["platlib"]:
        # The wheel has platform dependent code (is not pure) but
        # does not contain any extension module (does not
        # distribute any file in {platlib}) thus use generic
        # implementation and ABI tags.
        return _tags.Tag("py3", "none", None)

    return _tags.Tag(None, _compute_stable_abi(install_plan, limited_api), None)


def _aslist(s: T.Union[None, str, T.List[str]]) -> T.List[str]:
    if s is None:
        return []
    elif isinstance(s, str):
        return [s]
    return s


def _rmdashes(d: T.Optional[T.Dict[str, T.Any]]) -> T.Dict[str, T.Any]:
    if d is None:
        return {}

    return {k.replace("-", "_"): v for k, v in d.items()}


class MesonBuildHook(BuildHookInterface):

    PLUGIN_NAME = "meson"

    def initialize(self, version: str, build_data: T.Dict[str, T.Any]) -> None:
        # meson is not invoked when building an sdist
        if self.target_name != "wheel":
            return

        try:
            self._hook_config = parse_input(_rmdashes(self.config), HatchMesonConfig)
        except Exception as e:
            raise ConfigError(
                "[tool.hatch.build.hooks.meson] has incorrect configuration"
            ) from e

        self._setup_excludes()

        try:
            raw_config_settings = _rmdashes(_get_cmdline_config_settings())
            config_settings = parse_input(raw_config_settings, CmdlineConfig)
        except Exception as e:
            raise ConfigError("incorrect config settings passed to hatch-meson") from e

        # meson arguments from the command line take precedence over
        # arguments from the configuration file thus are added later
        self._hook_config.args.setup.extend(_aslist(config_settings.setup_args))
        self._hook_config.args.compile.extend(_aslist(config_settings.compile_args))
        self._hook_config.args.install.extend(_aslist(config_settings.install_args))

        self._source_dir = pathlib.Path(self.root)
        if config_settings.build_dir is not None:
            self._build_dir = pathlib.Path(config_settings.build_dir)
        else:
            self._build_dir = self._source_dir / "build" / _tags.get_abi_tag()

        self._meson_native_file = self._build_dir / "hatch-meson-native-file.ini"
        self._meson_cross_file = self._build_dir / "hatch-meson-cross-file.ini"

        self._meson = _get_meson_command(self._hook_config.meson)

        self._ninja = _env_ninja_command()
        if self._ninja is None:
            raise ConfigError(
                f"Could not find ninja version {_NINJA_REQUIRED_VERSION} or newer."
            )
        os.environ.setdefault("NINJA", self._ninja)

        self._build_dir.mkdir(parents=True, exist_ok=True)

        if _create_macos_crossfile(self._meson_cross_file):
            self._hook_config.args.setup += [
                "--cross-file",
                os.fspath(self._meson_cross_file),
            ]

        # write the native file
        native_file_entries = [f"python = '{sys.executable}'"]
        pkgconf = _pkgconf_binary_for_native_file()
        if pkgconf is not None:
            native_file_entries.append(f"pkg-config = '{pkgconf}'")

        native_file_data = "[binaries]\n" + "\n".join(native_file_entries) + "\n"
        self._meson_native_file.write_text(native_file_data, encoding="utf-8")

        # reconfigure if we have a valid Meson build directory. Meson
        # uses the presence of the 'meson-private/coredata.dat' file
        # in the build directory as indication that the build
        # directory has already been configured and arranges this file
        # to be created as late as possible or deleted if something
        # goes wrong during setup.
        reconfigure = (self._build_dir / "meson-private" / "coredata.dat").is_file()
        self._run_configure(reconfigure)

        # limited API
        self._limited_api = self._hook_config.limited_api
        if self._limited_api:
            # check whether limited API is disabled for the Meson project
            options = self._info("intro-buildoptions")
            value = next(
                (
                    option["value"]
                    for option in options
                    if option["name"] == "python.allow_limited_api"
                ),
                None,
            )
            if not value:
                self._limited_api = False

        if self._limited_api and bool(sysconfig.get_config_var("Py_GIL_DISABLED")):
            raise BuildError(
                "The package targets Python's Limited API, which is not supported by free-threaded CPython. "
                'The "python.allow_limited_api" Meson build option may be used to override the package default.'
            )

        # Time to build
        self._run_build()

        # Take anything that meson would have installed, and make sure that hatchling
        # will also install it
        install_plan = self._get_meson_install_plan()
        artifacts: T.List[str] = []

        # Scripts are just handled by hatchling
        for dst, src in install_plan["scripts"]:
            build_data["shared_scripts"][src] = dst.as_posix()

        # As is data
        for dst, src in install_plan["datadir"]:
            build_data["shared_data"][src] = dst.as_posix()

        # platlib, purelib are copied to the build tree in the package source path
        try:
            package_sources = list(self.build_config.sources)
        except ValueError:
            package_sources = []

        force_include = build_data["force_include"]
        if package_sources:
            pkgsrc: pathlib.Path = self._source_dir / package_sources[0]
        else:
            pkgsrc = self._source_dir

        # Only allow either purelib or platlib, not both
        if install_plan["purelib"] and install_plan["platlib"]:
            raise BuildError(
                "The install plan contains both purelib and platlib components, a "
                "'pure: false' argument may be missing in meson.build. "
                "It is recommended to set it in \"import('python').find_installation()\""
            )

        # copy purelib/platlib artifacts
        # - this is the desired behavior in editable mode as it allows other
        #   hatchling build hooks to do things with the files
        # - when making a real wheel, we could just always map it in via force_include,
        #   but if the user tries to make a wheel from a directory previously
        #   used in editable mode then hatchling may create a zipfile with two
        #   items of the same name in them... so just copy the file
        # - but if the user didn't tell hatchling which packages to use, if something
        #   was added to purelib/platlib then we add it to force_include

        for dstrel, src in install_plan["purelib"] + install_plan["platlib"]:
            dst = pkgsrc / dstrel

            # Anything added to force_include will be added to the wheel, even
            # in editable mode, which is not what we want
            if version != "editable":
                force_include[str(dst)] = dstrel.as_posix()

            artifacts.append(dst.as_posix())

            if str(dst) == str(src):
                continue

            dst.parent.mkdir(parents=True, exist_ok=True)

            # Use copy2 to preserve file mtime, which ensures incremental build works
            # if another project is consuming these files
            shutil.copy2(src, dst)

        tag = _compute_tag(install_plan, self._limited_api)

        is_pure = _install_is_pure(install_plan)
        if not is_pure:
            build_data["pure_python"] = False

        build_data["tag"] = str(tag)
        build_data["artifacts"].extend(artifacts)

    def clean(self, versions: T.List[str]) -> None:
        # TODO
        pass

    def dependencies(self) -> T.List[str]:
        return _get_requires_for_build_wheel()

    def _setup_excludes(self):
        to_exclude = ["meson.build"]

        # Feels like a hack, but it works? https://github.com/pypa/hatch/issues/1787
        # - should we exclude common source files by default, or make the user specify
        #   them? My gut says let the user do it, otherwise we have to provide an option
        #   to override it.
        build_config = self.build_config.build_config
        wheel_config = build_config.get("targets", {}).get("wheel", {})
        if "exclude" in wheel_config:
            wheel_config["exclude"] += to_exclude
        elif "exclude" in build_config:
            build_config["exclude"] += to_exclude
        else:
            build_config["exclude"] = to_exclude

    def _run_configure(self, reconfigure: bool = False) -> None:
        """Configure Meson project."""

        # TODO: only run this when configure args have changed
        setup_args = [
            os.fspath(self._source_dir),
            os.fspath(self._build_dir),
            # default build options
            "-Dbuildtype=release",
            "-Db_ndebug=if-release",
            "-Db_vscrt=md",
            # user build options
            *self._hook_config.args.setup,
            # pass native file last to have it override the python
            # interpreter path that may have been specified in user
            # provided native files
            f"--native-file={os.fspath(self._meson_native_file)}",
        ]
        if reconfigure:
            setup_args.insert(0, "--reconfigure")
        self._run(self._meson + ["setup", *setup_args])

    @property
    def _build_command(self) -> T.List[str]:
        assert self._ninja is not None  # help mypy out
        if sys.platform == "win32":
            # On Windows use 'meson compile' to setup the MSVC compiler
            # environment. Using the --ninja-args option allows to
            # provide the exact same semantics for the compile arguments
            # provided by the users.
            cmd = self._meson + ["compile"]
            args = self._hook_config.args.compile
            if args:
                cmd.append(f"--ninja-args={args!r}")
            return cmd
        return [self._ninja, *self._hook_config.args.compile]

    @functools.lru_cache(maxsize=None)
    def _run_build(self) -> None:
        """Build the Meson project."""
        self._run(self._build_command)

    @functools.lru_cache()
    def _info(self, name: str) -> T.Any:
        """Read info from meson-info directory."""
        info = self._build_dir.joinpath("meson-info", f"{name}.json")
        return json.loads(info.read_text(encoding="utf-8"))

    def _get_meson_install_plan(
        self,
    ) -> T.DefaultDict[str, T.List[T.Tuple[pathlib.Path, str]]]:
        """The files to be added to the wheel, organized by wheel path."""

        # Obtain the list of files Meson would install.
        install_plan = self._info("intro-install_plan")

        # Parse the 'meson install' args to extract --tags and --skip-subprojects
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("--tags")
        parser.add_argument("--skip-subprojects", nargs="?", const="*", default="")
        args, _ = parser.parse_known_args(self._hook_config.args.install)
        install_tags = {t.strip() for t in args.tags.split(",")} if args.tags else None
        skip_subprojects = {
            p for p in (p.strip() for p in args.skip_subprojects.split(",")) if p
        }

        # Filter the install plan accordingly.
        sources: T.DefaultDict[str, T.Dict[str, T.Dict[str, str]]] = (
            collections.defaultdict(dict)
        )
        for key, targets in install_plan.items():
            for target, details in targets.items():
                if install_tags is not None and details["tag"] not in install_tags:
                    continue
                subproject = details.get("subproject")
                if subproject is not None and (
                    subproject in skip_subprojects or "*" in skip_subprojects
                ):
                    continue
                sources[key][target] = details

        # Map Meson installation locations to wheel paths.
        return _map_to_wheel(sources)

    def _run(self, cmd: T.Sequence[str]):
        # Flush the line to ensure that the log line with the executed
        # command line appears before the command output. Without it,
        # the lines appear in the wrong order in pip output.
        self.app.display_info(f"+ {' '.join(cmd)}")
        r = subprocess.run(cmd, cwd=self._build_dir)
        if r.returncode != 0:
            raise SystemExit(r.returncode)


def _get_meson_projectinfo(cwd):
    meson = _get_meson_command(None)
    result = subprocess.run(
        meson + ["introspect", "meson.build", "--projectinfo"],
        cwd=cwd,
        text=True,
        capture_output=True,
    )
    if result.returncode != 0:
        raise BuildError(f"meson introspect failed: {result.stdout}")

    return json.loads(result.stdout)


class MesonMetadataHook(MetadataHookInterface):
    PLUGIN_NAME = "meson"

    def update(self, metadata: T.Dict):
        md = _get_meson_projectinfo(self.root)
        metadata["name"] = md["descriptive_name"]


class MesonVersionSource(VersionSourceInterface):
    PLUGIN_NAME = "meson"

    def get_version_data(self):
        md = _get_meson_projectinfo(self.root)
        if md["version"] == "undefined":
            raise BuildError("version not set for project() in meson.build")

        return {"version": md["version"]}


def _get_requires_for_build_wheel() -> T.List[str]:
    dependencies = []

    if os.environ.get("NINJA") is None and _env_ninja_command() is None:
        dependencies.append(f"ninja >= {_NINJA_REQUIRED_VERSION}")

    return dependencies


def _parse_version_string(string: str) -> T.Tuple[int, ...]:
    """Parse version string."""
    try:
        return tuple(map(int, string.split(".")[:3]))
    except ValueError:
        return (0,)


def _get_meson_command(
    meson: T.Optional[str] = None, *, version: str = _MESON_REQUIRED_VERSION
) -> T.List[str]:
    """Return the command to invoke meson."""

    # The MESON env var, if set, overrides the config value from pyproject.toml.
    # The config value, if given, is an absolute path or the name of an executable.
    meson = os.environ.get("MESON", meson or "meson")

    # If the specified Meson string ends in `.py`, we run it with the current
    # Python executable. This avoids problems for users on Windows, where
    # making a script executable isn't enough to get it to run when invoked
    # directly. For packages that vendor a forked Meson, the `meson.py` in the
    # root of the Meson repo can be used this way.
    if meson.endswith(".py"):
        if not os.path.exists(meson):
            raise ConfigError(f'Could not find the specified meson: "{meson}"')
        cmd = [sys.executable, os.path.abspath(meson)]
    else:
        cmd = [meson]

    # The meson Python package is a dependency of the hatch-meson Python
    # package, however, it may occur that the meson Python package is installed
    # but the corresponding meson command is not available in $PATH. Implement
    # a runtime check to verify that the build environment is setup correcly.
    try:
        r = subprocess.run(cmd + ["--version"], text=True, capture_output=True)
    except FileNotFoundError as err:
        raise ConfigError(f'meson executable "{meson}" not found') from err
    if r.returncode != 0:
        raise ConfigError(f"Could not execute meson: {r.stderr.strip()}")
    meson_version = r.stdout.strip()

    if _parse_version_string(meson_version) < _parse_version_string(version):
        raise ConfigError(
            f"Could not find meson version {version} or newer, found {meson_version}."
        )

    return cmd


def _env_ninja_command(*, version: str = _NINJA_REQUIRED_VERSION) -> T.Optional[str]:
    """Returns the path to ninja, or None if no ninja found."""
    required_version = _parse_version_string(version)
    env_ninja = os.environ.get("NINJA")
    ninja_candidates = [env_ninja] if env_ninja else ["ninja", "ninja-build", "samu"]
    for ninja in ninja_candidates:
        ninja_path = shutil.which(ninja)
        if ninja_path is not None:
            version = subprocess.run(
                [ninja_path, "--version"], check=False, text=True, capture_output=True
            ).stdout
            if _parse_version_string(version) >= required_version:
                return ninja_path
    return None


def _create_macos_crossfile(crossfile_path: pathlib.Path) -> bool:
    """setuptools-like ARCHFLAGS environment variable support"""
    if sysconfig.get_platform().startswith("macosx-"):
        archflags = os.environ.get("ARCHFLAGS", "").strip()
        if archflags:

            # parse the ARCHFLAGS environment variable
            parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
            parser.add_argument("-arch", action="append")
            args, unknown = parser.parse_known_args(archflags.split())
            if unknown:
                raise ConfigError(f"Unknown flag specified in $ARCHFLAGS={archflags!r}")
            arch, *other = set(args.arch)
            if other:
                raise ConfigError(
                    f"Multi-architecture builds are not supported but $ARCHFLAGS={archflags!r}"
                )

            macver, _, nativearch = platform.mac_ver()
            if arch != nativearch:
                x = os.environ.setdefault(
                    "_PYTHON_HOST_PLATFORM", f"macosx-{macver}-{arch}"
                )
                if not x.endswith(arch):
                    raise ConfigError(
                        f"$ARCHFLAGS={archflags!r} and $_PYTHON_HOST_PLATFORM={x!r} do not agree"
                    )
                family = "aarch64" if arch == "arm64" else arch
                cross_file_data = textwrap.dedent(f"""
                    [binaries]
                    c = ['cc', '-arch', {arch!r}]
                    cpp = ['c++', '-arch', {arch!r}]
                    objc = ['cc', '-arch', {arch!r}]
                    objcpp = ['c++', '-arch', {arch!r}]
                    strip = ['strip', '-arch', {arch!r}]
                    [host_machine]
                    system = 'darwin'
                    cpu = {arch!r}
                    cpu_family = {family!r}
                    endian = 'little'
                """)
                crossfile_path.write_text(cross_file_data, encoding="utf-8")
                return True

        # Support iOS targets. iOS does not have native build tools and always
        # requires cross compilation: synthesize the appropriate cross file.
        elif sysconfig.get_platform().startswith("ios-"):
            ios_ver = platform.ios_ver()  # type: ignore[attr-defined]

            arch = platform.machine()
            family = "aarch64" if arch == "arm64" else arch
            subsystem = "ios-simulator" if ios_ver.is_simulator else "ios"

            cross_file_data = textwrap.dedent(f"""
                [binaries]
                ar = '{arch}-apple-{subsystem}-ar'
                c = '{arch}-apple-{subsystem}-clang'
                cpp = '{arch}-apple-{subsystem}-clang++'
                objc = '{arch}-apple-{subsystem}-clang'
                objcpp = '{arch}-apple-{subsystem}-clang++'
                strip = '{arch}-apple-{subsystem}-strip'

                [host_machine]
                system = 'ios'
                subsystem = {subsystem!r}
                cpu = {arch!r}
                cpu_family = {family!r}
                endian = 'little'
            """)
            crossfile_path.write_text(cross_file_data, encoding="utf-8")
            return True

    return False



# SPDX-FileCopyrightText: 2022 The meson-python developers
#
# SPDX-License-Identifier: MIT

import pathlib
import shutil
import subprocess
import sys

import pytest

import hatchling.build

from .conftest import chdir

examples_dir = pathlib.Path(__file__).parent.parent / "docs" / "examples"


def test_spam(venv, tmp_path):
    """Test that the wheel for the example builds, installs, and imports."""
    project_path = tmp_path / "project"
    shutil.copytree(examples_dir / "spam", project_path)
    dist_path = project_path / "dist"
    with chdir(project_path):
        if sys.version_info < (3, 8):
            # The test project requires Python >= 3.8.
            with pytest.raises(SystemExit):
                hatchling.build.build_wheel(dist_path)
        else:
            wheel = hatchling.build.build_wheel(dist_path)
            subprocess.run(
                [venv.executable, "-m", "pip", "install", dist_path / wheel], check=True
            )
            output = subprocess.run(
                [venv.executable, "-c", "import spam; print(spam.add(1, 2))"],
                check=True,
                stdout=subprocess.PIPE,
            ).stdout
            assert int(output) == 3



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import importlib.machinery
import os
import pathlib
import platform
import sys
import sysconfig

from collections import defaultdict
from unittest.mock import Mock

import packaging.tags
import pytest

import hatch_meson.plugin
import hatch_meson._tags

from .conftest import adjust_packaging_platform_tag

# Test against the wheel tag generated by packaging module.
tag = next(packaging.tags.sys_tags())
ABI = tag.abi
INTERPRETER = tag.interpreter
PLATFORM = adjust_packaging_platform_tag(tag.platform)


def get_abi3_suffix():
    for suffix in importlib.machinery.EXTENSION_SUFFIXES:
        if ".abi3" in suffix:  # Unix
            return suffix
        elif suffix == ".pyd":  # Windows
            return suffix


SUFFIX = sysconfig.get_config_var("EXT_SUFFIX")
ABI3SUFFIX = get_abi3_suffix()


def test_wheel_tag():
    assert str(hatch_meson._tags.Tag()) == f"{INTERPRETER}-{ABI}-{PLATFORM}"
    assert str(hatch_meson._tags.Tag(abi="abi3")) == f"{INTERPRETER}-abi3-{PLATFORM}"


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_macos_platform_tag(monkeypatch):
    for minor in range(9, 16):
        monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", f"10.{minor}")
        version = (10, minor) if platform.mac_ver()[2] != "arm64" else (11, 0)
        assert (
            next(packaging.tags.mac_platforms(version))
            == hatch_meson._tags.get_platform_tag()
        )
    for major in range(11, 20):
        for minor in range(3):
            monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", f"{major}.{minor}")
            assert (
                next(packaging.tags.mac_platforms((major, minor)))
                == hatch_meson._tags.get_platform_tag()
            )
    for major in range(11, 13):
        monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", f"{major}.0")
        assert (
            next(packaging.tags.mac_platforms((major, 0)))
            == hatch_meson._tags.get_platform_tag()
        )
        monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", f"{major}")
        assert (
            next(packaging.tags.mac_platforms((major, 0)))
            == hatch_meson._tags.get_platform_tag()
        )


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_macos_platform_tag_arm64(monkeypatch):
    monkeypatch.setenv("_PYTHON_HOST_PLATFORM", "macosx-12.0-arm64")
    # Verify that the minimum platform ABI version on arm64 is 11.0.
    monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", "10.12")
    assert hatch_meson._tags.get_platform_tag() == "macosx_11_0_arm64"
    monkeypatch.setenv("MACOSX_DEPLOYMENT_TARGET", "12.34")
    assert hatch_meson._tags.get_platform_tag() == "macosx_12_0_arm64"


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_python_host_platform(monkeypatch):
    monkeypatch.setenv("_PYTHON_HOST_PLATFORM", "macosx-12.0-arm64")
    assert hatch_meson._tags.get_platform_tag().endswith("arm64")
    monkeypatch.setenv("_PYTHON_HOST_PLATFORM", "macosx-11.1-x86_64")
    assert hatch_meson._tags.get_platform_tag().endswith("x86_64")


@pytest.mark.skipif(sys.version_info < (3, 13), reason="requires Python 3.13 or higher")
@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_ios_platform_tag(monkeypatch):
    # Mock being on iOS
    monkeypatch.setattr(sys.implementation, "_multiarch", "arm64-iphoneos")
    monkeypatch.setattr(
        sysconfig, "get_platform", Mock(return_value="ios-13.0-arm64-iphoneos")
    )
    ios_ver = platform.IOSVersionInfo("iOS", "13.0", "iPhone", False)
    monkeypatch.setattr(platform, "ios_ver", Mock(return_value=ios_ver))

    # Check the default value
    assert (
        next(packaging.tags.ios_platforms((13, 0)))
        == hatch_meson._tags.get_platform_tag()
    )

    # Check the value when IPHONEOS_DEPLOYMENT_TARGET is set.
    for major in range(13, 20):
        for minor in range(3):
            monkeypatch.setenv("IPHONEOS_DEPLOYMENT_TARGET", f"{major}.{minor}")
            assert (
                next(packaging.tags.ios_platforms((major, minor)))
                == hatch_meson._tags.get_platform_tag()
            )


def wheel_builder_test_factory(content, pure=True, limited_api=False):
    manifest = defaultdict(list)
    manifest.update(
        {
            key: [(pathlib.Path(x), os.path.join("build", x)) for x in value]
            for key, value in content.items()
        }
    )
    return hatch_meson.plugin._compute_tag(manifest, limited_api)


def test_tag_empty_wheel():
    tag = wheel_builder_test_factory({})
    assert str(tag) == "py3-none-any"


def test_tag_purelib_wheel():
    tag = wheel_builder_test_factory(
        {
            "purelib": ["pure.py"],
        }
    )
    assert str(tag) == "py3-none-any"


def test_tag_platlib_wheel():
    tag = wheel_builder_test_factory(
        {
            "platlib": [f"extension{SUFFIX}"],
        }
    )
    assert str(tag) == f"{INTERPRETER}-{ABI}-{PLATFORM}"


def test_tag_stable_abi():
    tag = wheel_builder_test_factory(
        {
            "platlib": [f"extension{ABI3SUFFIX}"],
        },
        limited_api=True,
    )
    # PyPy does not support the stable ABI.
    abi = "abi3" if "__pypy__" not in sys.builtin_module_names else ABI
    assert str(tag) == f"{INTERPRETER}-{abi}-{PLATFORM}"


@pytest.mark.xfail(
    sys.version_info < (3, 8) and sys.platform == "win32",
    reason="Extension modules suffix without ABI tags",
)
@pytest.mark.xfail(
    "__pypy__" in sys.builtin_module_names,
    reason="PyPy does not support the stable ABI",
)
def test_tag_mixed_abi():

    with pytest.raises(
        hatch_meson.plugin.BuildError,
        match="The package declares compatibility with Python limited API but ",
    ):
        tag = wheel_builder_test_factory(
            {
                "platlib": [f"extension{ABI3SUFFIX}", f"another{SUFFIX}"],
            },
            pure=False,
            limited_api=True,
        )

        assert str(tag) == f"{INTERPRETER}-abi3-{PLATFORM}"



from hatchling.plugin import hookimpl

from .plugin import MesonBuildHook, MesonMetadataHook, MesonVersionSource


@hookimpl
def hatch_register_build_hook():
    return MesonBuildHook


@hookimpl
def hatch_register_metadata_hook():
    return MesonMetadataHook


@hookimpl
def hatch_register_version_source():
    return MesonVersionSource



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import os
import re
import shutil
import stat
import subprocess
import sys
import sysconfig

import hatch_meson.plugin

import hatchling.build
import packaging.tags
import pytest
import wheel.wheelfile

from .conftest import adjust_packaging_platform_tag

_meson_ver_str = subprocess.run(
    ["meson", "--version"], check=True, stdout=subprocess.PIPE, text=True
).stdout
MESON_VERSION = tuple(map(int, _meson_ver_str.split(".")[:3]))

EXT_SUFFIX = sysconfig.get_config_var("EXT_SUFFIX")
if sys.version_info <= (3, 8, 7):
    if MESON_VERSION >= (0, 99):
        # Fixed in Meson 1.0, see https://github.com/mesonbuild/meson/pull/10961.
        from distutils.sysconfig import get_config_var

        EXT_SUFFIX = get_config_var("EXT_SUFFIX")

if sys.platform in {"win32", "cygwin"}:
    EXT_IMP_SUFFIX = re.sub(
        r".(pyd|dll)$", ".lib" if shutil.which("cl.exe") else ".dll.a", EXT_SUFFIX
    )

LIB_SUFFIX = {
    "cygwin": ".dll",
    "darwin": ".dylib",
    "win32": ".dll",
}.get(sys.platform, ".so")

NOGIL_BUILD = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))

# Test against the wheel tag generated by packaging module.
tag = next(packaging.tags.sys_tags())
ABI = tag.abi
INTERPRETER = tag.interpreter
PLATFORM = adjust_packaging_platform_tag(tag.platform)


def wheel_contents(artifact):
    # Sometimes directories have entries, sometimes not, so we filter them out.
    return {entry for entry in artifact.namelist() if not entry.endswith("/")}


def test_scipy_like(wheel_scipy_like):
    # This test is meant to exercise features commonly needed by a regular
    # Python package for scientific computing or data science:
    #   - C and Cython extensions,
    #   - including generated code,
    #   - using `install_subdir`,
    #   - packaging data files with extensions not known to Meson
    artifact = wheel.wheelfile.WheelFile(wheel_scipy_like)

    expecting = {
        "mypkg-2.3.4.dist-info/METADATA",
        "mypkg-2.3.4.dist-info/RECORD",
        "mypkg-2.3.4.dist-info/WHEEL",
        "mypkg/__init__.py",
        "mypkg/__config__.py",
        f"mypkg/extmod{EXT_SUFFIX}",
        f"mypkg/cy_extmod{EXT_SUFFIX}",
        "mypkg/submod/__init__.py",
        "mypkg/submod/unknown_filetype.npq",
    }
    if sys.platform in {"win32", "cygwin"}:
        # Currently Meson is installing .dll.a (import libraries) next
        # to .pyd extension modules. Those are very small, so it's not
        # a major issue - just sloppy. Ensure we don't fail on those.
        expecting.update(
            {
                f"mypkg/extmod{EXT_IMP_SUFFIX}",
                f"mypkg/cy_extmod{EXT_IMP_SUFFIX}",
            }
        )
    assert wheel_contents(artifact) == expecting

    name = artifact.parsed_filename
    assert name is not None
    assert name.group("pyver") == INTERPRETER
    assert name.group("abi") == ABI
    assert name.group("plat") == PLATFORM


# hatch-meson doesn't support mixed purelib and platlib
# def test_purelib_and_platlib(wheel_purelib_and_platlib):
#     artifact = wheel.wheelfile.WheelFile(wheel_purelib_and_platlib)
#
#     expecting = {
#         f"plat{EXT_SUFFIX}",
#         "purelib_and_platlib-1.0.0.data/purelib/pure.py",
#         "pure.py",
#         "purelib_and_platlib-1.0.0.dist-info/METADATA",
#         "purelib_and_platlib-1.0.0.dist-info/RECORD",
#         "purelib_and_platlib-1.0.0.dist-info/WHEEL",
#     }
#     if sys.platform in {"win32", "cygwin"}:
#         # Currently Meson is installing .dll.a (import libraries) next
#         # to .pyd extension modules. Those are very small, so it's not
#         # a major issue - just sloppy. Ensure we don't fail on those.
#         expecting.update({f"plat{EXT_IMP_SUFFIX}"})
#
#     assert wheel_contents(artifact) == expecting


def test_purelib_and_platlib(copyof_purelib_and_platlib, tmp_path):
    with pytest.raises(
        hatch_meson.plugin.BuildError,
        match="The install plan contains both purelib and platlib",
    ):
        hatchling.build.build_wheel(tmp_path)


def test_pure(wheel_pure):
    artifact = wheel.wheelfile.WheelFile(wheel_pure)

    assert wheel_contents(artifact) == {
        "pure-1.0.0.dist-info/METADATA",
        "pure-1.0.0.dist-info/RECORD",
        "pure-1.0.0.dist-info/WHEEL",
        "pure.py",
    }


def test_configure_data(wheel_configure_data):
    artifact = wheel.wheelfile.WheelFile(wheel_configure_data)

    assert wheel_contents(artifact) == {
        "configure_data.py",
        "configure_data-1.0.0.dist-info/METADATA",
        "configure_data-1.0.0.dist-info/RECORD",
        "configure_data-1.0.0.dist-info/WHEEL",
    }


#
# TODO: add these tests back in when bundling is restored
#

# @pytest.mark.skipif(
#     sys.platform not in {"linux", "darwin"}, reason="Not supported on this platform"
# )
# def test_contents(package_library, wheel_library):
#     artifact = wheel.wheelfile.WheelFile(wheel_library)
#
#     assert wheel_contents(artifact) == {
#         f".library.mesonpy.libs/libexample{LIB_SUFFIX}",
#         "library-1.0.0.data/headers/examplelib.h",
#         "library-1.0.0.data/scripts/example",
#         "library-1.0.0.dist-info/METADATA",
#         "library-1.0.0.dist-info/RECORD",
#         "library-1.0.0.dist-info/WHEEL",
#     }


# @pytest.mark.skipif(
#     sys.platform not in {"linux", "darwin"}, reason="Not supported on this platform"
# )
# def test_local_lib(venv, wheel_link_against_local_lib):
#     venv.pip("install", wheel_link_against_local_lib)
#     output = venv.python("-c", "import example; print(example.example_sum(1, 2))")
#     assert int(output) == 3


def test_sharedlib_in_package(venv, wheel_sharedlib_in_package):
    venv.pip("install", wheel_sharedlib_in_package)
    output = venv.python("-c", "import mypkg; print(mypkg.example_sum(2, 5))")
    assert int(output) == 7
    output = venv.python("-c", "import mypkg; print(mypkg.example_prod(6, 7))")
    assert int(output) == 42


@pytest.mark.skipif(MESON_VERSION < (1, 3, 0), reason="Meson version too old")
def test_link_library_in_subproject(venv, wheel_link_library_in_subproject):
    venv.pip("install", wheel_link_library_in_subproject)
    output = venv.python("-c", "import foo; print(foo.example_sum(3, 6))")
    assert int(output) == 9


# @pytest.mark.skipif(
#     sys.platform not in {"linux", "darwin"}, reason="Not supported on this platform"
# )
# def test_rpath(wheel_link_against_local_lib, tmp_path):
#     artifact = wheel.wheelfile.WheelFile(wheel_link_against_local_lib)
#     artifact.extractall(tmp_path)
#
#     origin = {"linux": "$ORIGIN", "darwin": "@loader_path"}[sys.platform]
#     expected = {
#         f"{origin}/.link_against_local_lib.mesonpy.libs",
#         "custom-rpath",
#     }
#
#     rpath = set(mesonpy._rpath._get_rpath(tmp_path / f"example{EXT_SUFFIX}"))
#     # Verify that rpath is a superset of the expected one: linking to
#     # the Python runtime may require additional rpath entries.
#     assert rpath >= expected


# @pytest.mark.skipif(
#     sys.platform not in {"linux", "darwin"}, reason="Not supported on this platform"
# )
# def test_uneeded_rpath(wheel_plat, tmp_path):
#     artifact = wheel.wheelfile.WheelFile(wheel_plat)
#     artifact.extractall(tmp_path)
#
#     origin = {"linux": "$ORIGIN", "darwin": "@loader_path"}[sys.platform]
#
#     rpath = mesonpy._rpath._get_rpath(tmp_path / f"plat{EXT_SUFFIX}")
#     for path in rpath:
#         assert origin not in path


@pytest.mark.skipif(
    sys.platform not in {"linux", "darwin"}, reason="Not supported on this platform"
)
def test_executable_bit(wheel_executable_bit):
    artifact = wheel.wheelfile.WheelFile(wheel_executable_bit)

    executable_files = {
        "executable_module.py",
        "executable_bit-1.0.0.data/scripts/example",
        "executable_bit-1.0.0.data/scripts/example-script",
    }
    for info in artifact.infolist():
        mode = (info.external_attr >> 16) & 0o777
        assert bool(mode & stat.S_IXUSR) == (info.filename in executable_files)


def test_detect_wheel_tag_module(wheel_plat):
    name = wheel.wheelfile.WheelFile(wheel_plat).parsed_filename
    assert name is not None
    assert name.group("pyver") == INTERPRETER
    assert name.group("abi") == ABI
    assert name.group("plat") == PLATFORM


def test_detect_wheel_tag_script(wheel_executable):
    name = wheel.wheelfile.WheelFile(wheel_executable).parsed_filename
    assert name is not None
    assert name.group("pyver") == "py3"
    assert name.group("abi") == "none"
    assert name.group("plat") == PLATFORM


# hatchling does entry points
# def test_entrypoints(wheel_full_metadata):
#     artifact = wheel.wheelfile.WheelFile(wheel_full_metadata)

#     with artifact.open("full_metadata-1.2.3.dist-info/entry_points.txt") as f:
#         assert (
#             f.read().decode().strip()
#             == textwrap.dedent(
#                 """
#             [something.custom]
#             example = example:custom

#             [console_scripts]
#             example-cli = example:cli

#             [gui_scripts]
#             example-gui = example:gui
#         """
#             ).strip()
#         )


# Hatchling deals with editable details
#
# def test_top_level_modules(package_module_types):
#     with mesonpy._project() as project:
#         builder = mesonpy._EditableWheelBuilder(
#             project._metadata, project._manifest, project._limited_api
#         )
#         assert set(builder._top_level_modules) == {
#             "file",
#             "package",
#             "namespace",
#             "native",
#         }


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
@pytest.mark.parametrize("arch", ["x86_64", "arm64"])
def test_archflags_envvar(copyof_plat, monkeypatch, tmp_path, arch):
    try:
        monkeypatch.setenv("ARCHFLAGS", f"-arch {arch}")
        filename = hatchling.build.build_wheel(tmp_path)
        name = wheel.wheelfile.WheelFile(tmp_path / filename).parsed_filename
        assert name is not None
        assert name.group("plat").endswith(arch)
    finally:
        # revert environment variable setting done by the in-process build
        os.environ.pop("_PYTHON_HOST_PLATFORM", None)


def test_subprojects(wheel_subproject):
    artifact = wheel.wheelfile.WheelFile(wheel_subproject)
    assert wheel_contents(artifact) == {
        "subproject-1.0.0.dist-info/METADATA",
        "subproject-1.0.0.dist-info/RECORD",
        "subproject-1.0.0.dist-info/WHEEL",
        "subproject.py",
        "dep.py",
    }


# Requires Meson 1.2.0, see https://github.com/mesonbuild/meson/pull/11909.
@pytest.mark.skipif(MESON_VERSION < (1, 1, 99), reason="Meson version too old")
@pytest.mark.parametrize(("arg"), ["--skip-subprojects", "--skip-subprojects=dep"])
def test_skip_subprojects(copyof_subproject, tmp_path, arg):
    filename = hatchling.build.build_wheel(tmp_path, {"install-args": [arg]})
    artifact = wheel.wheelfile.WheelFile(tmp_path / filename)
    assert wheel_contents(artifact) == {
        "subproject-1.0.0.dist-info/METADATA",
        "subproject-1.0.0.dist-info/RECORD",
        "subproject-1.0.0.dist-info/WHEEL",
        "subproject.py",
    }


# Requires Meson 1.3.0, see https://github.com/mesonbuild/meson/pull/11745.
@pytest.mark.skipif(MESON_VERSION < (1, 2, 99), reason="Meson version too old")
@pytest.mark.skipif(
    NOGIL_BUILD, reason="Free-threaded CPython does not support the limited API"
)
@pytest.mark.xfail(
    "__pypy__" in sys.builtin_module_names,
    reason="PyPy does not support the abi3 platform tag for wheels",
)
def test_limited_api(wheel_limited_api):
    artifact = wheel.wheelfile.WheelFile(wheel_limited_api)
    name = artifact.parsed_filename
    assert name is not None
    assert name.group("pyver") == INTERPRETER
    assert name.group("abi") == "abi3"
    assert name.group("plat") == PLATFORM


# Requires Meson 1.3.0, see https://github.com/mesonbuild/meson/pull/11745.
@pytest.mark.skipif(MESON_VERSION < (1, 2, 99), reason="Meson version too old")
@pytest.mark.skipif(
    NOGIL_BUILD, reason="Free-threaded CPython does not support the limited API"
)
@pytest.mark.xfail(
    "__pypy__" in sys.builtin_module_names,
    reason="PyPy does not use special modules suffix for stable ABI",
)
def test_limited_api_bad(copyof_limited_api, tmp_path):
    with pytest.raises(
        hatch_meson.plugin.BuildError,
        match="The package declares compatibility with Python limited API but ",
    ):
        hatchling.build.build_wheel(tmp_path, {"setup-args": ["-Dextra=true"]})


# Requires Meson 1.3.0, see https://github.com/mesonbuild/meson/pull/11745.
@pytest.mark.skipif(MESON_VERSION < (1, 2, 99), reason="Meson version too old")
def test_limited_api_disabled(copyof_limited_api, tmp_path):
    filename = hatchling.build.build_wheel(
        tmp_path, {"setup-args": ["-Dpython.allow_limited_api=false"]}
    )
    artifact = wheel.wheelfile.WheelFile(tmp_path / filename)
    name = artifact.parsed_filename
    assert name is not None
    assert name.group("pyver") == INTERPRETER
    assert name.group("abi") == ABI
    assert name.group("plat") == PLATFORM


def test_install_subdir(wheel_install_subdir):
    artifact = wheel.wheelfile.WheelFile(wheel_install_subdir)
    # Handling of the exclude_files and exclude_directories requires
    # Meson 1.1.0, see https://github.com/mesonbuild/meson/pull/11432.
    # Run the test anyway to ensure that meson-python can produce a
    # wheel also for older versions of Meson.
    if MESON_VERSION >= (1, 1, 99):
        assert wheel_contents(artifact) == {
            "install_subdir-1.0.0.dist-info/METADATA",
            "install_subdir-1.0.0.dist-info/RECORD",
            "install_subdir-1.0.0.dist-info/WHEEL",
            "subdir/__init__.py",
            "subdir/test.py",
            "test/module.py",
            "nested/deep/deep.py",
            "nested/nested.py",
        }


def test_vendored_meson(wheel_vendored_meson):
    # This test will error if the vendored meson.py wrapper script in
    # the test package isn't used.
    pass


def test_encoding(wheel_encoding):
    artifact = wheel.wheelfile.WheelFile(wheel_encoding)
    assert wheel_contents(artifact) == {
        "encoding-1.0.0.dist-info/METADATA",
        "encoding-1.0.0.dist-info/RECORD",
        "encoding-1.0.0.dist-info/WHEEL",
        "テスト.py",
    }


def test_custom_target_install_dir(wheel_custom_target_dir):
    artifact = wheel.wheelfile.WheelFile(wheel_custom_target_dir)
    assert wheel_contents(artifact) == {
        "custom_target_dir-1.0.0.dist-info/METADATA",
        "custom_target_dir-1.0.0.dist-info/RECORD",
        "custom_target_dir-1.0.0.dist-info/WHEEL",
        "package/generated/one.py",
        "package/generated/two.py",
    }


def test_custom_target_install_dir_editable(tmp_path, copyof_custom_target_dir):
    filename = hatchling.build.build_editable(tmp_path)
    artifact = wheel.wheelfile.WheelFile(tmp_path / filename)
    assert wheel_contents(artifact) == {
        "_custom_target_dir.pth",
        "custom_target_dir-1.0.0.dist-info/METADATA",
        "custom_target_dir-1.0.0.dist-info/RECORD",
        "custom_target_dir-1.0.0.dist-info/WHEEL",
    }


def test_install_tags(copyof_install_tags, tmp_path):
    filename = hatchling.build.build_wheel(
        tmp_path, {"install-args": ["--tags", "pkg1"]}
    )
    artifact = wheel.wheelfile.WheelFile(tmp_path / filename)
    assert wheel_contents(artifact) == {
        "install_tags-1.0.0.dist-info/METADATA",
        "install_tags-1.0.0.dist-info/RECORD",
        "install_tags-1.0.0.dist-info/WHEEL",
        "pkg1.py",
    }



# SPDX-FileCopyrightText: 2024 The meson-python developers
#
# SPDX-License-Identifier: MIT

import os
import sys


# start-literalinclude
def _append_to_sharedlib_load_path():
    """
    Ensure the shared libraries in this package can be loaded on Windows.

    Windows lacks a concept equivalent to RPATH: Python extension modules
    cannot find DLLs installed outside the DLL search path. This function
    ensures that the location of the shared libraries distributed inside this
    Python package is in the DLL search path of the process.

    The Windows DLL search path includes the object depending on it is located:
    the DLL search path needs to be augmented only when the Python extension
    modules and the DLLs they require are installed in separate directories.
    Cygwin does not have the same default library search path: all locations
    where the shared libraries are installed need to be added to the search
    path.

    This function is very similar to the snippet inserted into the main
    ``__init__.py`` of a package by ``delvewheel`` when it vendors external
    shared libraries.

    .. note::

        `os.add_dll_directory` is only available for Python 3.8 and later, and
        in the Conda ``python`` packages it works as advertised only for
        version 3.10 and later. For older Python versions, pre-loading the DLLs
        with `ctypes.WinDLL` may be preferred.
    """
    basedir = os.path.dirname(__file__)
    subdir = os.path.join(basedir, "sub")
    if os.name == "nt":
        os.add_dll_directory(subdir)
    elif sys.platform == "cygwin":
        os.environ["PATH"] = os.pathsep.join((os.environ["PATH"], basedir, subdir))


_append_to_sharedlib_load_path()
# end-literalinclude


from ._example import example_prod, example_sum  # noqa: E402

__all__ = ["example_prod", "example_sum"]



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import stat
import sys
import tarfile
import textwrap

import pytest

from .conftest import metadata


def test_dynamic_version(sdist_dynamic_version):
    with tarfile.open(sdist_dynamic_version, "r:gz") as sdist:
        sdist_pkg_info = sdist.extractfile("dynamic_version-1.0.0/PKG-INFO").read()

    sdist_metadata = metadata(sdist_pkg_info)
    metadata_version = sdist_metadata.get("metadata_version", "2.3")

    expected_metadata = metadata(textwrap.dedent(f"""\
            Metadata-Version: {metadata_version}
            Name: dynamic-version
            Version: 1.0.0
        """))

    assert sdist_metadata == expected_metadata


# def test_contents(sdist_library):
#     with tarfile.open(sdist_library, "r:gz") as sdist:
#         names = {member.name for member in sdist.getmembers()}
#         mtimes = {member.mtime for member in sdist.getmembers()}

#     assert names == {
#         "library-1.0.0/example.c",
#         "library-1.0.0/examplelib.c",
#         "library-1.0.0/examplelib.h",
#         "library-1.0.0/meson.build",
#         "library-1.0.0/pyproject.toml",
#         "library-1.0.0/PKG-INFO",
#     }

#     # All the archive members have a valid mtime.
#     assert 0 not in mtimes


def test_contents_subdirs(sdist_subdirs):
    with tarfile.open(sdist_subdirs, "r:gz") as sdist:
        names = {member.name for member in sdist.getmembers()}
        mtimes = {member.mtime for member in sdist.getmembers()}

    assert names == {
        "subdirs-1.0.0/PKG-INFO",
        "subdirs-1.0.0/meson.build",
        "subdirs-1.0.0/pyproject.toml",
        "subdirs-1.0.0/subdirs/__init__.py",
        "subdirs-1.0.0/subdirs/a/__init__.py",
        "subdirs-1.0.0/subdirs/a/b/c.py",
        "subdirs-1.0.0/subdirs/b/c.py",
    }

    # All the archive members have a valid mtime.
    assert 0 not in mtimes


# This tests something that hatchling handles
# def test_contents_unstaged(package_pure, tmp_path):
#     new = textwrap.dedent(
#         """
#         def bar():
#             return 'foo'
#     """
#     ).strip()

#     old = pathlib.Path("pure.py").read_text()

#     with in_git_repo_context():
#         try:
#             pathlib.Path("pure.py").write_text(new)
#             pathlib.Path("other.py").touch()
#             sdist_path = hatchling.build.build_sdist(tmp_path)
#         finally:
#             pathlib.Path("pure.py").write_text(old)
#             pathlib.Path("other.py").unlink()

#     with tarfile.open(tmp_path / sdist_path, "r:gz") as sdist:
#         names = {member.name for member in sdist.getmembers()}
#         mtimes = {member.mtime for member in sdist.getmembers()}
#         data = sdist.extractfile("pure-1.0.0/pure.py").read().replace(b"\r\n", b"\n")

#     # Verify that uncommitted changes are not included in the sdist.
#     assert names == {
#         "pure-1.0.0/PKG-INFO",
#         "pure-1.0.0/meson.build",
#         "pure-1.0.0/pure.py",
#         "pure-1.0.0/pyproject.toml",
#     }
#     assert data == old.encode()

#     # All the archive members have a valid mtime.
#     assert 0 not in mtimes


@pytest.mark.skipif(
    sys.platform in {"win32", "cygwin"},
    reason="Platform does not support executable bit",
)
def test_executable_bit(sdist_executable_bit):
    expected = {
        "executable_bit-1.0.0/PKG-INFO": False,
        "executable_bit-1.0.0/example-script.py": True,
        "executable_bit-1.0.0/example.c": False,
        "executable_bit-1.0.0/executable_module.py": True,
        "executable_bit-1.0.0/meson.build": False,
        "executable_bit-1.0.0/pyproject.toml": False,
    }

    with tarfile.open(sdist_executable_bit, "r:gz") as sdist:
        for member in sdist.getmembers():
            assert bool(member.mode & stat.S_IXUSR) == expected[member.name]


# This meson-python test will fail because there isn't a way to run a user's
# dist script without running `meson dist`, and we rely on hatchling to create
# the sdist instead.
#
# def test_generated_files(sdist_generated_files):
#     with tarfile.open(sdist_generated_files, "r:gz") as sdist:
#         names = {member.name for member in sdist.getmembers()}
#         mtimes = {member.mtime for member in sdist.getmembers()}
#
#     assert names == {
#         "executable_bit-1.0.0/PKG-INFO",
#         "executable_bit-1.0.0/example-script.py",
#         "executable_bit-1.0.0/example.c",
#         "executable_bit-1.0.0/executable_module.py",
#         "executable_bit-1.0.0/meson.build",
#         "executable_bit-1.0.0/pyproject.toml",
#         "executable_bit-1.0.0/_version_meson.py",
#         "executable_bit-1.0.0/generate_version.py",
#     }
#
#     # All the archive members have a valid mtime.
#     assert 0 not in mtimes


# def test_reproducible(package_pure, tmp_path):
#     with in_git_repo_context():
#         t1 = time.time()
#         sdist_path_a = hatchling.build.build_sdist(tmp_path / "a")
#         t2 = time.time()
#         # Ensure that the two sdists are build at least one second apart.
#         time.sleep(max(t1 + 1.0 - t2, 0.0))
#         sdist_path_b = hatchling.build.build_sdist(tmp_path / "b")

#     assert sdist_path_a == sdist_path_b
#     assert (
#         tmp_path.joinpath("a", sdist_path_a).read_bytes()
#         == tmp_path.joinpath("b", sdist_path_b).read_bytes()
#     )



# SPDX-FileCopyrightText: 2021 The meson-python developers
#
# SPDX-License-Identifier: MIT

import ast
import importlib.metadata
import os
import shutil
import sys

import hatch_meson.plugin

import hatchling.build
import pytest

from .conftest import in_git_repo_context, package_dir


def test_missing_dynamic_version(copyof_missing_dynamic_version, tmp_path):
    with pytest.raises(
        hatch_meson.plugin.BuildError,
        match="version not set for project()",
    ):
        hatchling.build.build_wheel(tmp_path)


def test_user_args(copyof_user_args, tmp_path, monkeypatch):
    project_run = hatch_meson.plugin.MesonBuildHook._run
    cmds = []
    args = []

    def wrapper(self, cmd):
        # intercept and filter out test arguments and forward the call
        if cmd[:2] == ["meson", "compile"]:
            # when using meson compile instead of ninja directly, the
            # arguments needs to be unmarshalled from the form used to
            # pass them to the --ninja-args option
            assert cmd[-1].startswith("--ninja-args=")
            cmds.append(cmd[:2])
            args.append(ast.literal_eval(cmd[-1].split("=")[1]))
        elif cmd[:1] == ["meson"]:
            cmds.append(cmd[:2])
            args.append(cmd[2:])
        else:
            # direct ninja invocation
            cmds.append([os.path.basename(cmd[0])])
            args.append(cmd[1:])
        return project_run(
            self,
            [x for x in cmd if not x.startswith(("config-", "cli-", "--ninja-args"))],
        )

    monkeypatch.setattr(hatch_meson.plugin.MesonBuildHook, "_run", wrapper)

    config_settings = {
        "setup-args": ["cli-setup"],
        "compile-args": ["cli-compile"],
        "install-args": ["cli-install"],
    }

    with in_git_repo_context():
        hatchling.build.build_wheel(tmp_path, config_settings)

    # check that the right commands are executed, namely that 'meson
    # compile' is used on Windows rather than a 'ninja' direct
    # invocation.
    assert cmds == [
        # wheel: calls to 'meson setup', 'meson compile', and 'meson install'
        ["meson", "setup"],
        ["meson", "compile"] if sys.platform == "win32" else ["ninja"],
    ]

    # check that the user options are passed to the invoked commands
    expected = [
        # sdist: calls to 'meson setup' and 'meson dist'
        # ["config-setup", "cli-setup"],
        # ["config-dist", "cli-dist"],
        # wheel: calls to 'meson setup', 'meson compile', and 'meson install'
        ["config-setup", "cli-setup"],
        ["config-compile", "cli-compile"],
        ["config-install", "cli-install"],
    ]
    for expected_args, cmd_args in zip(expected, args):
        for arg in expected_args:
            assert arg in cmd_args


def test_unknown_user_args_meson_args(copyof_unknown_user_args_meson_args, tmp_path):
    with pytest.raises(hatch_meson.plugin.ConfigError):
        hatchling.build.build_wheel(tmp_path)


def test_native_file_uses_pkg_config_pypi_when_pkgconf_package_installed(
    copyof_pure, tmp_path, monkeypatch
):
    native_file_data = None
    project_run = hatch_meson.plugin.MesonBuildHook._run
    script_path = tmp_path / "venv-bin" / "pkgconf-pypi"
    script_path.parent.mkdir()
    script_path.write_text("#!/bin/sh\n", encoding="utf-8")

    class FakeDistribution:
        files = [importlib.metadata.PackagePath("../../../bin/pkgconf-pypi")]

        def locate_file(self, path):
            assert str(path) == "../../../bin/pkgconf-pypi"
            return script_path

    metadata_distribution = importlib.metadata.distribution

    def distribution(name):
        if name == "pkgconf":
            return FakeDistribution()
        return metadata_distribution(name)

    def wrapper(self, cmd):
        nonlocal native_file_data
        native_file_data = self._meson_native_file.read_text(encoding="utf-8")
        return project_run(self, cmd)

    monkeypatch.setattr(importlib.metadata, "distribution", distribution)
    monkeypatch.setattr(hatch_meson.plugin.MesonBuildHook, "_run", wrapper)

    with in_git_repo_context():
        hatchling.build.build_wheel(tmp_path)

    assert native_file_data is not None
    assert "python = '" in native_file_data
    assert f"pkg-config = '{script_path}'" in native_file_data


# test_install_tags moved to test_wheel.py


@pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason="unittest.mock doesn't support the required APIs for this test",
)
def test_invalid_build_dir(copyof_pure, tmp_path, mocker):
    meson = mocker.spy(hatch_meson.plugin.MesonBuildHook, "_run")
    build_dir = copyof_pure / "build"
    config_settings = {"build-dir": str(build_dir)}

    # configure the project
    hatchling.build.build_wheel(tmp_path, config_settings)
    assert len(meson.call_args_list) == 2
    assert meson.call_args_list[0].args[1][1] == "setup"
    assert "--reconfigure" not in meson.call_args_list[0].args[1]
    meson.reset_mock()

    # subsequent builds with the same build directory result in a setup --reconfigure
    hatchling.build.build_wheel(tmp_path, config_settings)
    assert len(meson.call_args_list) == 2
    assert meson.call_args_list[0].args[1][1] == "setup"
    assert "--reconfigure" in meson.call_args_list[0].args[1]
    meson.reset_mock()

    # corrupting the build direcory setup is run again
    build_dir.joinpath("meson-private/coredata.dat").unlink()
    hatchling.build.build_wheel(tmp_path, config_settings)
    assert len(meson.call_args_list) == 2
    assert meson.call_args_list[0].args[1][1] == "setup"
    assert "--reconfigure" not in meson.call_args_list[0].args[1]
    meson.reset_mock()

    # removing the build directory things should still work
    shutil.rmtree(build_dir)
    hatchling.build.build_wheel(tmp_path, config_settings)
    assert len(meson.call_args_list) == 2
    assert meson.call_args_list[0].args[1][1] == "setup"
    assert "--reconfigure" not in meson.call_args_list[0].args[1]


@pytest.mark.skipif(
    not os.getenv("CI") or sys.platform != "win32", reason="Requires MSVC"
)
def test_compiler(venv, copyof_detect_compiler, tmp_path):
    # Check that things are setup properly to use the MSVC compiler on
    # Windows. This effectively means running the compilation step
    # with 'meson compile' instead of 'ninja' on Windows. Run this
    # test only on CI where we know that MSVC is available.
    wheel = hatchling.build.build_wheel(tmp_path, {"setup-args": ["--vsenv"]})
    venv.pip("install", os.fspath(tmp_path / wheel))
    compiler = venv.python(
        "-c", "import detect_compiler; print(detect_compiler.compiler())"
    ).strip()
    assert compiler == "msvc"


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
@pytest.mark.parametrize(
    "archflags",
    [
        "-arch x86_64",
        "-arch arm64",
        "-arch arm64 -arch arm64",
    ],
)
def test_archflags_envvar_parsing(copyof_plat, monkeypatch, archflags, tmp_path):
    try:
        monkeypatch.setenv("ARCHFLAGS", archflags)
        arch = archflags.split()[-1]
        whl = hatchling.build.build_wheel(tmp_path)
        assert whl.endswith(f"_{arch}.whl")
    finally:
        # revert environment variable setting done by the in-process build
        os.environ.pop("_PYTHON_HOST_PLATFORM", None)


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
@pytest.mark.parametrize(
    "archflags",
    [
        "-arch arm64 -arch x86_64",
        "-arch arm64 -DFOO=1",
    ],
)
def test_archflags_envvar_parsing_invalid(
    copyof_plat, monkeypatch, archflags, tmp_path
):
    try:
        monkeypatch.setenv("ARCHFLAGS", archflags)
        with pytest.raises(hatch_meson.plugin.ConfigError):
            hatchling.build.build_wheel(tmp_path)
    finally:
        # revert environment variable setting done by the in-process build
        os.environ.pop("_PYTHON_HOST_PLATFORM", None)


# @pytest.mark.skipif(sys.version_info < (3, 13), reason="requires Python 3.13 or higher")
# @pytest.mark.parametrize(
#     "multiarch",
#     [
#         "arm64-iphoneos",
#         "arm64-iphonesimulator",
#         "x86_64-iphonesimulator",
#     ],
# )
# def test_ios_project(package_simple, monkeypatch, multiarch, tmp_path):
#     arch, abi = multiarch.split("-")
#     subsystem = "ios-simulator" if abi == "iphonesimulator" else "ios"

#     # Mock being on iOS
#     monkeypatch.setattr(sys, "platform", "ios")
#     monkeypatch.setattr(platform, "machine", Mock(return_value=arch))
#     monkeypatch.setattr(
#         sysconfig, "get_platform", Mock(return_value=f"ios-13.0-{multiarch}")
#     )
#     ios_ver = platform.IOSVersionInfo(
#         "iOS", "13.0", "iPhone", multiarch.endswith("simulator")
#     )
#     monkeypatch.setattr(platform, "ios_ver", Mock(return_value=ios_ver))

#     # Create an iOS project.
#     project = mesonpy.Project(source_dir=package_simple, build_dir=tmp_path)

#     # Meson configuration points at the cross file
#     assert project._meson_args["setup"] == [
#         "--cross-file",
#         os.fspath(tmp_path / "meson-python-cross-file.ini"),
#     ]

#     # Meson config files exist, and have some relevant keys
#     assert (tmp_path / "meson-python-native-file.ini").exists()
#     assert (tmp_path / "meson-python-cross-file.ini").exists()

#     cross_config = (tmp_path / "meson-python-cross-file.ini").read_text()

#     assert "\nsystem = 'ios'\n" in cross_config
#     assert f"\nc = '{arch}-apple-{subsystem}-clang'\n" in cross_config
#     assert f"\nsubsystem = '{subsystem}'\n" in cross_config



# SPDX-FileCopyrightText: 2023 The meson-python developers
#
# SPDX-License-Identifier: MIT

import sys

from mesonbuild import mesonmain

if "setup" in sys.argv:
    sys.argv.append("-Dcustom-meson-used=true")


if __name__ == "__main__":
    sys.exit(mesonmain.main())
